# 3D Camera motion
# ================
#
# With some limitations, Ren'Py can simulate 3D camera motion. This will easily
# let you achieve a parallax effect by placing sprites and assets on a 3D field.
# Ren'Py applies transforms to each layers which are registered as 3D layer by
# positions of a camera and 3D layers on a 3D field to simulate 3D camera motion.
#
# If you use this feature, in the first, set additional layers to be
# registered as 3D layer.
#
# For example, Write a code like below in options.rpy and add 'background', 'middle', 'forward'  layers.::
#
#         config.layers = ['master', 'background', 'middle', 'forward', 'transient', 'screens', 'overlay']
#
# Second, register layers which participate to a 3D motion as 3d layers by :func:`register_3d_layer`::
#
#         init python:
#             register_3d_layer('background', 'middle', 'forward')
#
# Then, a camera and layers can move. ::
#
#          label start:
#              # reset a camera and layers positions
#              $ camera_reset()
#              scene bg onlayer background
#              show A onlayer middle
#              show B onlayer forward
#              # It takes 0 second to move layers
#              $ layer_move("background", 2000)
#              $ layer_move("middle", 1500)
#              $ layer_move("forward", 1000)
#              with dissolve
#              'It takes 1 second to move a camera to (1800, 0, 0)'
#              $ camera_move(1800, 0, 0, 0, 1)
#              'It takes 5 seconds to move a camera to (0, 0, 1600)'
#              $ camera_move(0, 0, 1600, 0, 5)
#              'A camera moves to (0, 0, 0) at the moment'
#              $ camera_move(0, 0, 0)
#              'It takes 1 second to rotate a camera to 180 degrees'
#              $ camera_move(0, 0, 0, 180, 1)
#              'It takes 1 second to rotate a camera to -180 degrees and 0.5 second to move a camera to (-1800, 0, 500)'
#              $ camera_moves( ( (0, 0, 0, 0, 1, 'linear'), (-1800, 0, 500, 0, 1.5, 'linear') ) )
#              'a camera shuttles between (-1800, 0, 500) and (0, 0, 0)'
#              $ camera_moves( ( (0, 0, 0, 0, .5, 'linear'), (-1800, 0, 500, 0, 1, 'linear') ), loop=True)
#
# When :var:`config.developer` is True, pressing position_viewer (by default,
# "shift+P"), will open Position Viewer. This allow you to adjustment a camera
# ,3D layers and images positions by bars and manual inputs and see the result at
# the moment.
#
# Notice that 3D camera motion has some limits.:
#
# * 3D camera motion applies transforms to 3D layers, so the show layer statement or
#   :func:`renpy.show_layer_at` to 3D layers can't be usable.
#
# * z coordinates of 3D layers don't affect the stacking order of 3d layers.
#
# * A camera and a layer can't move at the same time.
#
# * By default, the scene, hide statements use master or the given layer
#   only. If you use 3D layers preferentially, set Ren'py like below.::
#
#         init -1 python hide:
#             def hide(name, layer='master'):
#                 for l in _3d_layers:
#                     if renpy.showing(name, l):
#                         renpy.hide(name, l)
#                         break
#                 else:
#                     renpy.hide(name, layer)
#
#             config.hide = hide
#
#             def scene(layer='master'):
#
#                 renpy.scene(layer)
#                 for l in _3d_layers:
#                     renpy.scene(l)
#
#             config.scene = scene
#
# Camera Functions
# ----------------
#
#    def register_3d_layer(*layers):
#        """
#         :doc: camera
#
#         Register layers as 3D layers. 3D layers is applied transforms to by
#         positions of a camera and 3D layers. You can use this in game again
#         and again.
#
#         `layers`
#              This should be the string or strings of a layer name.
#         """
#
#    def camera_reset():
#        """
#         :doc: camera
#
#         Reset a camera and 3D layers positions.
#         """
#
#    def camera_move(x, y, z, rotate=0, duration=0, warper='linear', subpixel=True, loop=False):
#        """
#         :doc: camera
#
#         Move the coordinate and rotate of a camera and apply transforms to all 3D layers.
#
#         `x`
#              the x coordinate of a camera
#         `y`
#              the y coordinate of a camera
#         `z`
#              the z coordinate of a camera
#         `rotate`
#              Defaul 0, the rotate of a camera
#         `duration`
#              Default 0, this is the second times taken to move a camera.
#         `warper`
#              Default 'linear', this should be string and the name of a warper
#              registered with ATL.
#         `subpixel`
#              Default True, if True, causes things to be drawn on the screen
#              using subpixel positioning
#         `loop`
#              Default False, if True, this motion repeats.
#         """
#
#    def layer_move(layer, z, duration=0, warper='linear', subpixel=True, loop=False):
#        """
#         :doc: camera
#
#         Move the z coordinate of a layer and apply transform to the layer.
#
#         `layer`
#              the string of a layer name to be moved
#         `z`
#              the z coordinate of a layer
#         `duration`
#              Default 0, this is the second times taken to move a camera.
#         `warper`
#              Default 'linear', this should be the string of the name of a
#              warper registered with ATL.
#         `subpixel`
#              Default True, if True, causes things to be drawn on the screen
#              using subpixel positioning
#         `loop`
#              Default False, if True, this motion repeats.
#         """
#
#    def camera_moves(check_points, loop=False, subpixel=True):
#        """
#         :doc: camera
#
#         Move a camera through check points and apply transforms to all 3D
#         layers.
#
#         `check_points`
#              A tuple of tuples of x, y, z, rotate, duration, warper.
#         `loop`
#              Default False, if True, this sequence of motions repeats.
#         `subpixel`
#              Default True, if True, causes things to be drawn on the screen
#              using subpixel positioning
#         """
#
#    def layer_moves(layer, check_points, loop=False, subpixel=True):
#        """
#         :doc: camera
#
#         Move a layer through check points and apply transform to the layer.
#
#         `layer`
#              the string of a layer name to be moved
#         `check_points`
#              A tuple of tuples of z, duration, warper
#         `loop`
#              Default False, if True, this sequence of motions repeats.
#         `subpixel`
#              Default True, if True, causes things to be drawn on the screen
#              using subpixel positioning
#         """
# """
# 3D Camera Motion Simulator 
#
# Ren'Py にカメラの移動を再現する関数を追加します。これにより奥行きの表現が簡単
# に可能になります。この機能を利用するには、先ず3Dレイヤーに登録するレイヤーを
# 追加しなければなりません。
#
# 例
#
# options.rpy に以下のような設定を加え、'background', 'middle', 'forward' レイヤーを追加します。
#
#        config.layers = [ 'master', 'background', 'middle', 'forward', 'transient', 'screens', 'overlay']
#
# 次に register_3d_layer 関数で3Dモーションに参加するレイヤーを3Dレイヤーとして登録します。
#
#    init python:
#        register_3d_layer( 'background', 'middle', 'forward',)
#
# 以上の設定で、ゲーム中でカメラ動かせるようになります。
#
#    label start:
#        # カメラ、レイヤー位置をリセットします。
#        $ camera_reset()
#        # 0秒かけてレイヤー位置を変更します。
#        $ layer_move("background", 2000)
#        $ layer_move("middle", 1500)
#        $ layer_move("forward", 1000)
#        scene bg onlayer background
#        show A onlayer middle
#        show B onlayer forward
#        '1秒かけてカメラを(0, 130, 0)まで移動させます。'
#        $ camera_move(0, 130, 0, 0, 1)
#        '5秒かけてカメラを(-70, 130, 1000)まで移動させます。'
#        $ camera_move(-70, 130, 1000, 0, 5)
#        '0秒かけてカメラを(0, 0, 0)まで移動させます。'
#        $ camera_move(0, 0, 0)
#        '1秒かけてカメラを時計回りに180度回転します。'
#        $ camera_move(0, 0, 0, 180, 1)
#        '1秒かけてカメラを反時計回りに180度回転させ、4秒かけてカメラを(100, 100, 100)まで移動させます。'
#        $ camera_moves( ( (0, 0, 0, 0, 1, 'linear'), (100, 100, 100, 0, 5, 'linear') ) )
#        'カメラが(100, 100, 100)から(0, 0, 0)までを繰り返し往復します'
#        $ camera_moves( ( (0, 0, 0, 0, 1, 'linear'), (100, 100, 100, 0, 2, 'linear') ), loop=True)
#
# Shift+Pでカメラビューワーも利用可能です。カメラビューワーではバーからカメラと
# レイヤー位置を変更し、その結果を目視で確認できます。Shift + Tで利用可能なトラ
# ンスフォームビューワーも助けになるでしょう。上記2つのビューワーは
# config.developer が True の時のみ利用出来ます。
#
# カメラ機能にはいくつかの制限があることに注意してください。
# カメラモーションは3Dレイヤーに対してトランスフォームを適用して再現しているた
# め、3レイヤーに対する show layer ステートメントや renpy.show_layer_at 関数は使
# 用出来ません。
#
# カメラとレイヤーを同時に動かすことは出来ません。
#
# レイヤーの重なり順はレイヤーのz座標を反映しません。
#
# デフォルトではRen'Pyの scene, show, hide ステートメントは master レイヤー又は
# 指定されたレイヤーにのみ作用します。3Dレイヤーに対して優先的に作用させたい場
# 合は以下のようにしてRen'Pyの設定を変更してください。
#        init -1 python hide:
#
#            def hide(name, layer='master'):
#                for l in _3d_layers:
#                    if renpy.showing(name, l):
#                        renpy.hide(name, l)
#                else:
#                    renpy.hide(name, layer)
#
#            config.hide = hide
#
#            def scene(layer='master'):
#
#                renpy.scene(layer)
#                for l in _3d_layers:
#                    renpy.scene(l)
#
#            config.scene = scene
# """
#
#     def register_3d_layer(*layers):
#         """
#          :doc: camera
#
#          レイヤーを3Dレイヤーとして登録します。3Dレイヤーに登録されたレイヤーに
#          はカメラの動きに基づいたトランスフォームが適用されます。設定はゲーム中
#          に何度でも変更可能です。
#
#          `layers`
#               3Dレイヤーに登録する、一つ以上のレイヤー名の文字列です。
#          """
#
#     def camera_reset():
#         """
#          :doc: camera
#
#          カメラとレイヤーを初期位置に戻します。
#          """
#
#     def set_film_size(film_size):
#         """
#          :doc: camera
#
#          カメラのフィルムサイズを変更し、カメラ位置をリセットします。
#
#          `film_size`
#               カメラのフィルムサイズの数字です。
#          """
#
#     def camera_move(x, y, z, rotate=0, duration=0, warper='linear', subpixel=True, loop=False):
#         """
#          :doc: camera
#
#          カメラの座標と角度を指定の時間かけて変更します。カメラの移動に伴い、各
#          3Dレイヤーにトランスフォームが適用されます。
#
#          `x`
#               カメラのx座標の数字です。
#          `y`
#               カメラのy座標の数字です。
#          `z`
#               カメラのy座標の数字です。
#          `rotate`
#               デフォルトは0で、カメラの傾きの数字です。
#          `duration`
#               デフォルトは0で、カメラ移動にかかる時間の秒数です。
#          `loop`
#               デフォルトは Falseで、True ならこのモーションが繰り替えされます。
#          `warper`
#               デフォルトは'linear'で, 時間補間に使用する関数名の文字列です。こ
#               れにはATLに登録された時間補間関数が指定出来ます。
#          `subpixel`
#               デフォルトは Trueで、 True なら、1 pixel 以下の値を使用して画面
#               に描画します。
#          """
#
#     def camera_moves(check_points, loop=False, subpixel=True):
#         """
#          :doc: camera
#
#          1つ以上のチェックポイントを通ってカメラを動かします。カメラの移動に伴い、各
#          3Dレイヤーにトランスフォームが適用されます。
#
#          `check_points`
#               カメラのx,y,z座標、傾き、カメラの移動開始からその位置に到達する
#               までの秒数、その間の時間補間関数名の文字列からなるタプルのタプ
#               ルです。カメラはここで指定された各タプルを順番に移動するので、
#               タプルは時系列に沿って並べてください。
#          `loop`
#               デフォルトは Falseで、True ならこのモーションが繰り替えされます。
#          `subpixel`
#               デフォルトは Trueで、 True なら、1 pixel 以下の値を使用して画面
#               に描画します。
#          """
#
#     def layer_move(layer, z, duration=0, warper='linear', subpixel=True, loop=False):
#         """
#          :doc: camera
#
#          レイヤーのz座標を指定の時間かけて変更します。レイヤーの移動に伴い、そ
#          のレイヤーにトランスフォームが適用されます。
#
#          `layer`
#               移動するレイヤー名の文字列です。
#          `z`
#               レイヤーの移動先z座標です。
#          `duration`
#               デフォルトは0で、レイヤー移動にかかる時間の秒数です。
#          `loop`
#               デフォルトは Falseで、True ならこのモーションが繰り替えされます。
#          `warper`
#               デフォルトは'linear'で, 時間補間に使用する関数名の文字列です。こ
#               れにはATLに登録された時間補間関数が指定出来ます。
#          `subpixel`
#               デフォルトは Trueで、 True なら、1 pixel 以下の値を使用して画面
#               に描画します。
#
#     def layer_moves(layer, check_points, loop=False, subpixel=True):
#         """
#          :doc: camera
#
#          1つ以上のチェックポイントを通って3Dレイヤーを動かします。レイヤーの移
#          動に伴い、そのレイヤーにトランスフォームが適用されます。
#
#          `layer`
#               移動するレイヤー名の文字列です。
#          `check_points`
#               レイヤーのz座標、レイヤーの移動開始からその位置に到達するまでの
#               秒数、その間の時間補間関数の文字列からなるタプルのタプルです。
#               レイヤーはここで指定された各タプルを順番に移動するので、タプル
#               は時系列に沿って並べてください。
#          `loop`
#               デフォルトは Falseで、True ならこのモーションが繰り替えされます。
#          `subpixel`
#               デフォルトは Trueで、 True なら、1 pixel 以下の値を使用して画面
#               に描画します。
#

init -1600 python:

    ##########################################################################
    # Camera functions

    # value of _3d_layers isn't included in rollback.
    # so I copy it many times.
    _3d_layers = {}
    _camera_x = 0
    _camera_y = 0
    _camera_z = 0
    _camera_rotate = 0
    _focal_length = 147.40
    _layer_z = 1848.9


    def register_3d_layer(*layers):
        """
         :doc: camera

         Register layers as 3D layers. 3D layers is applied transforms to by
         positions of a camera and 3D layers. You can use this in game again
         and again.

         `layers`
              This should be the string or strings of a layer name.
         """
        global _3d_layers
        af_3d_layers = {}
        for layer in layers:
            af_3d_layers[layer] = _layer_z
        _3d_layers = af_3d_layers

    def camera_reset():
        """
         :doc: camera

         Reset a camera and 3D layers positions.
         """
        camera_move(0, 0, 0, 0)
        for layer in _3d_layers:
            layer_move(layer, _layer_z)

    def camera_move(x, y, z, rotate=0, duration=0, warper='linear', subpixel=True, loop=False):
        """
         :doc: camera

         Move the coordinate and rotate of a camera and apply transforms to all 3D layers.

         `x`
              the x coordinate of a camera
         `y`
              the y coordinate of a camera
         `z`
              the z coordinate of a camera
         `rotate`
              Defaul 0, the rotate of a camera
         `duration`
              Default 0, this is the second times taken to move a camera.
         `warper`
              Default 'linear', this should be string and the name of a warper
              registered with ATL.
         `subpixel`
              Default True, if True, causes things to be drawn on the screen
              using subpixel positioning
         `loop`
              Default False, if True, this motion repeats.
         """

        camera_moves(((x, y, z, rotate, duration, warper), ), subpixel=subpixel, loop=loop)

    def layer_move(layer, z, duration=0, warper='linear', subpixel=True, loop=False):
        """
         :doc: camera

         Move the z coordinate of a layer and apply transform to the layer.

         `layer`
              the string of a layer name to be moved
         `z`
              the z coordinate of a layer
         `duration`
              Default 0, this is the second times taken to move a camera.
         `warper`
              Default 'linear', this should be the string of the name of a
              warper registered with ATL.
         `subpixel`
              Default True, if True, causes things to be drawn on the screen
              using subpixel positioning
         `loop`
              Default False, if True, this motion repeats.
         """

        layer_moves(layer, ((z, duration, warper), ), subpixel=subpixel, loop=loop)

    def camera_moves(check_points, loop=False, subpixel=True, play=True):
        """
         :doc: camera

         Move a camera through check points and apply transforms to all 3D
         layers.

         `check_points`
              A tuple of (x, y, z, rotate, duration, warper)
         `loop`
              Default False, if True, this sequence of motions repeats.
         `subpixel`
              Default True, if True, causes things to be drawn on the screen
              using subpixel positioning
         """
        global _camera_x, _camera_y, _camera_z, _camera_rotate

        af_3d_layers = {}
        for layer in _3d_layers:
            start_xanchor = _focal_length*_camera_x/(config.screen_width *_layer_z) + .5
            start_yanchor = _focal_length*_camera_y/(config.screen_height*_layer_z) + .5

            check_points2 = [(_camera_z, _camera_rotate, start_xanchor, start_yanchor, _3d_layers[layer], 0, None), ]
            for c in check_points:
                xanchor = _focal_length*c[0]/(config.screen_width *_layer_z) + .5
                yanchor = _focal_length*c[1]/(config.screen_height*_layer_z) + .5
                duration = float(c[4])
                check_points2.append((c[2], c[3], xanchor, yanchor, _3d_layers[layer], duration, c[5]))

            renpy.game.context().scene_lists.set_layer_at_list(layer, [Transform(function=renpy.curry(_camera_trans)(check_points=check_points2, loop=loop, subpixel=subpixel, play=play))])

        _camera_x = check_points[-1][0]
        _camera_y = check_points[-1][1]
        _camera_z = check_points[-1][2]
        _camera_rotate = check_points[-1][3]

    def layer_moves(layer, check_points, loop=False, subpixel=True):
        """
         :doc: camera

         Move a layer through check points and apply transform to the layer.

         `layer`
              the string of a layer name to be moved
         `check_points`
              A tuple of (z, duration, warper)
         `loop`
              Default False, if True, this sequence of motions repeats.
         `subpixel`
              Default True, if True, causes things to be drawn on the screen
              using subpixel positioning
         """
        global _3d_layers, _camera_x, _camera_y, _camera_z, _camera_rotate

        start_xanchor = _focal_length*_camera_x/(config.screen_width *_layer_z) + .5
        start_yanchor = _focal_length*_camera_y/(config.screen_height*_layer_z) + .5

        check_points2 = [(_camera_z, _camera_rotate, start_xanchor, start_yanchor, _3d_layers[layer], 0, None), ]
        af_3d_layers = _3d_layers.copy()
        for c in check_points:
            check_points2.append((_camera_z, _camera_rotate, start_xanchor, start_yanchor, c[0], float(c[1]), c[2]))

        renpy.game.context().scene_lists.set_layer_at_list(layer, [Transform(function=renpy.curry(_camera_trans)(check_points=check_points2, loop=loop, subpixel=subpixel))])

        af_3d_layers[layer] = check_points[-1][0]
        _3d_layers = af_3d_layers

    def _camera_trans(tran, st, at, check_points, loop, subpixel, play=True):
        # This is used when the time bar is changed by PositionViewer
        if not play:
            st = _viewers.time
        # check_points = (z, r, xanchor, yanchor, layer_z, duration, warper)
        duration = check_points[-1][5]
        tran.xpos    = .5 
        tran.ypos    = .5
        tran.subpixel = subpixel
        tran.transform_anchor = True
        if duration > 0:
            g = st/duration
            if loop:
                g = g % 1

            for i in xrange(1, len(check_points)):
                checkpoint_g = check_points[i][5]/duration
                pre_checkpoint_g = check_points[i-1][5]/duration
                if g <= checkpoint_g:
                    start = check_points[i-1]
                    goal = check_points[i]
                    if checkpoint_g != pre_checkpoint_g:
                        ch_g = renpy.atl.warpers[goal[6]]((g - pre_checkpoint_g) / (checkpoint_g - pre_checkpoint_g))
                    else:
                        ch_g = 1.

                    z = ch_g*(goal[0]-start[0])+start[0]
                    layer_z = ch_g*(goal[4]-start[4])+start[4]
                    distance = float(layer_z - z)
                    if distance == 0:
                        distance = .1

                    tran.xanchor = ch_g*(goal[2]-start[2]) + start[2]
                    tran.yanchor = ch_g*(goal[3]-start[3]) + start[3]
                    tran.rotate  = ch_g*(goal[1]  -  start[1]) + start[1]

                    if distance >= 0:
                        tran.alpha = 1
                        tran.zoom = _layer_z / distance
                    else:
                        tran.alpha = 0
                    return .005
        distance = float(check_points[-1][4] - check_points[-1][0])
        if distance == 0:
            distance = .1
        tran.xanchor = check_points[-1][2]
        tran.yanchor = check_points[-1][3]
        tran.zoom = _layer_z / distance
        tran.rotate = check_points[-1][1]
        if distance >= 0:
            tran.alpha = 1
        else:
            tran.alpha = 0
        return None



init -1600 python in _viewers:

    ##########################################################################
    #{(tag, layer):[(time, check_point, warper)...]}
    all_anchor_points = {}
    time = 0
    sorted_anchor_points = []
    warper = "linear"

    def edit_time():
        global time
        v = renpy.invoke_in_new_context(renpy.call_screen, "_input_screen")
        if v:
            try:
                v = renpy.python.py_eval(v)
                if v < 0:
                    return
                time = v
            except:
                renpy.notify(_("Please type value"))

    def next_time():
        for i, t in enumerate(sorted_anchor_points):
            if time < t:
                change_time(sorted_anchor_points[i])
                break
        else:
            change_time(sorted_anchor_points[0])
        return

    def prev_time():
        for i, t in enumerate(sorted_anchor_points):
            if time <= t:
                change_time(sorted_anchor_points[i-1])
                break
        else:
            change_time(sorted_anchor_points[-1])
        return

    def select_time_warper():
        renpy.invoke_in_new_context(renpy.call_screen, "_warper_selecter")

    def clear_anchor_points():
        all_anchor_points.clear()

    def remove_anchor_point(time):
        if time <= 0:
            return

        remove_list=[]
        for k, v in all_anchor_points.items():
            for t, c, w in v:
                if t == time:
                    remove_list.append( (k, (t, c, w)) )
        for k, a in remove_list:
            if len(all_anchor_points[k]) == 1:
                del all_anchor_points[k]
            else:
                all_anchor_points[k].remove(a)
        sort_anchor_points()

    def sort_anchor_points():
        global sorted_anchor_points
        sorted_anchor_points[:] = []
        for anchor_points in all_anchor_points.values():
            for t, c, w in anchor_points:
                sorted_anchor_points.append(t)
        sorted_anchor_points.sort()

    def put_clipboard():
        camera_check_points = []
        if "camera" in all_anchor_points:
            for k, c, w in all_anchor_points["camera"]:
                camera_check_points.append(c)

        string = ""
        for k, v in all_anchor_points.items():
            if k == "camera" and camera_check_points:
                if len(camera_check_points) == 1 and camera_check_points[0][0:5] == (renpy.store._camera_viewer._camera_x, renpy.store._camera_viewer._camera_y, renpy.store._camera_viewer._camera_z, renpy.store._camera_viewer._camera_rotate, 0,):
                    continue
                string += """
    $camera_moves(%s, loop=False, subpixel=True)""" % (camera_check_points)
            else:
                string += """
    show %s onlayer %s:
        subpixel True """ % k
                kwargs_org = renpy.store._transform_viewer.state_org[k[1]][k[0]].copy()
                for p in renpy.store._transform_viewer.props_default:
                    if kwargs_org[p] is None and p != "rotate":
                        kwargs_org[p] = renpy.store._transform_viewer.props_default[p][0]

                for prop, value in v[0][1].items():
                    if value != kwargs_org[prop]:
                        string += "%s %s " % (prop, value)
                for i, t in enumerate(v[1:]):
                    string += """
        %s %s """ % (t[2], t[0]-v[i][0])
                    for prop, value in t[1].items():
                    # for i, (prop, value) in enumerate(t[1].items()):
                    #     for t2, k2, w2, in v[:i]:
                    #         if value = [prop]:
                        string += "%s %s " % (prop, value)

        try:
            from pygame import scrap, locals
            scrap.put(locals.SCRAP_TEXT, string)
        except:
            renpy.notify(_("Can't open clipboard"))
        else:
            renpy.notify(_("Put strings on clipboard"))

    def change_time(v):
        global time
        time = v
        renpy.store._transform_viewer.play(False)
        renpy.store._camera_viewer.play(False)
        renpy.restart_interaction()

init -1600 python:
    ##########################################################################
    # TransformViewer
    class TransformViewer(object):
        def __init__(self):

            self.int_range = 1500
            self.float_range = 7.0
            # layer->tag->property->value
            self.state_org = {}
            # {property:(default, float)}, default is used when property can't be got.
            self.props_default = {
            "xpos":(0., False),
            "ypos":(0., False),
            "xanchor":(0., False),
            "yanchor":(0., False),
            "xzoom":(1., True),
            "yzoom":(1., True),
            "zoom":(1., True),
            "rotate":(0, False),
            "alpha":(1., True),
            "additive":(0., True),
            }

        def init(self):
            if not config.developer:
                return
            sle = renpy.game.context().scene_lists
            # back up for reset()
            for layer in config.layers:
                self.state_org[layer] = {}
                for tag in sle.layers[layer]:
                    d = sle.get_displayable_by_tag(layer, tag[0])
                    if isinstance(d, renpy.display.screen.ScreenDisplayable):
                        break
                    pos = renpy.get_placement(d)
                    state = getattr(d, "state", None)

                    self.state_org[layer][tag[0]] = {}
                    for p in ["xpos", "ypos", "xanchor", "yanchor"]:
                        self.state_org[layer][tag[0]][p] = getattr(pos, p, None)
                    for p in self.props_default:
                        if not self.state_org[layer][tag[0]].has_key(p):
                            self.state_org[layer][tag[0]][p] = getattr(state, p, None)

        def reset(self):
            for layer in config.layers:
                for tag, props in self.state_org[layer].iteritems():
                    if tag and props:
                        kwargs = props.copy()
                        for p in self.props_default:
                            if kwargs[p] is None and p != "rotate":
                                kwargs[p] = self.props_default[p][0]
                        renpy.show(tag, [Transform(**kwargs)], layer=layer)
            renpy.restart_interaction()

        def set_anchor_point(self, layer, tag, kwargs):

            anchor_points = _viewers.all_anchor_points.get((tag, layer), [])
            if anchor_points:
                for i, (t, c, w) in enumerate(anchor_points):
                    if _viewers.time < t:
                        anchor_points.insert(i, (_viewers.time, kwargs, _viewers.warper))
                        break
                    elif _viewers.time == t:
                        anchor_points[i] = (_viewers.time, kwargs, _viewers.warper)
                        break
                else:
                    anchor_points.append((_viewers.time, kwargs, _viewers.warper))
            else:
                if _viewers.time == 0:
                    _viewers.all_anchor_points[(tag, layer)] = [(_viewers.time, kwargs, _viewers.warper)]
                else:
                    kwargs_org = {}
                    for p in self.props_default:
                        kwargs_org[p] = self.get_state(layer, tag, p, False)
                    _viewers.all_anchor_points[(tag, layer)] = [(0, kwargs_org, _viewers.warper), (_viewers.time, kwargs, _viewers.warper)]
            _viewers.sort_anchor_points()

        def play(self, play):
            for layer in config.layers:
                for tag in self.state_org[layer]:
                    if (tag, layer) in _viewers.all_anchor_points:
                        renpy.show(tag, [Transform(function=renpy.curry(self.transform)(check_points=_viewers.all_anchor_points[(tag, layer)], play=play))], layer=layer)

        def transform(self, tran, st, at, check_points, play=True, subpixel=True):
            # check_points = [(time, kwargs, warper)..]
            time = check_points[-1][0]
            tran.subpixel = subpixel
            if not play:
                st = _viewers.time

            if st < time:
                g = st/time 

                for i in xrange(1, len(check_points)):
                    checkpoint_g = check_points[i][0]/time
                    pre_checkpoint_g = check_points[i-1][0]/time
                    if g <= checkpoint_g:
                        start = check_points[i-1][1]
                        goal = check_points[i][1]
                        if checkpoint_g != pre_checkpoint_g:
                            ch_g = renpy.atl.warpers[check_points[i][2]]((g - pre_checkpoint_g) / (checkpoint_g - pre_checkpoint_g))
                        else:
                            ch_g = 1.
                        for k in self.props_default:
                            if goal[k] is not None and start[k] is not None:
                                setattr(tran, k, ch_g*(goal[k]-start[k])+start[k])

                        return .005
            for k, v in check_points[-1][1].iteritems():
                setattr(tran, k, v)
            return None

        def generate_changed(self, layer, tag, prop, int=False):
            def changed(v):
                kwargs = {}
                for p in self.props_default:
                    kwargs[p] = self.get_state(layer, tag, p, False)
                if int and not self.props_default[prop][1]:
                    kwargs[prop] = v - self.int_range
                else:
                    kwargs[prop] = v -self.float_range

                self.set_anchor_point(layer, tag, kwargs)
                renpy.show(tag, [Transform(**kwargs)], layer=layer)
                renpy.restart_interaction()
            return changed

        def get_state(self, layer, tag, prop, default=True):
            sle = renpy.game.context().scene_lists
            if tag:
                d = sle.get_displayable_by_tag(layer, tag)
                pos = renpy.get_placement(d)
                state = getattr(pos, prop, None)
                if state is None:
                    state = getattr(getattr(d, "state", None), prop, None)
                # set default
                if state is None and default:
                    state = self.props_default[prop][0]
                if state and self.props_default[prop][1]:
                    state = float(state)
                return state
            return None

        def put_prop_clipboard(self, prop, value):
            try:
                from pygame import scrap, locals
                scrap.put(locals.SCRAP_TEXT, "%s %s" % (prop, value))
            except:
                renpy.notify(_("Can't open clipboard"))
            else:
                renpy.notify(__('Putted "%s %s" on clipboard') % (prop, value))

        def put_show_clipboard(self, tag, layer):
            string = "show %s onlayer %s" % (tag, layer)
            for k, v in self.props_default.iteritems():
                value = self.get_state(layer, tag, k)
                if value != v[0]:
                    if string.find(":") < 0:
                        string += ":\n    "
                    string += "%s %s " % (k, value)
            try:
                from pygame import scrap, locals
                scrap.put(locals.SCRAP_TEXT, string)
            except:
                renpy.notify(_("Can't open clipboard"))
            else:
                renpy.notify(__('Putted "%s" on clipboard') % string)

        def edit_value(self, function, int=False):
            v = renpy.invoke_in_new_context(renpy.call_screen, "_input_screen")
            if v:
                try:
                    if int:
                        v = renpy.python.py_eval(v) + self.int_range
                    else:
                        v = renpy.python.py_eval(v) + self.float_range
                    function(v)
                except:
                    renpy.notify(_("Please type value"))
    _transform_viewer = TransformViewer()

    ##########################################################################
    # CameraViewer
    class CameraViewer(object):

        def __init__(self):
            self.range_camera_pos   = 6000
            self.range_rotate       = 360
            self.range_layer_z   = 10000

        def init(self):
            if not config.developer:
                return
            self._camera_x = _camera_x
            self._camera_y = _camera_y
            self._camera_z = _camera_z
            self._camera_rotate = _camera_rotate
            self._3d_layers = _3d_layers.copy()
            self.set_anchor_point()

        def camera_reset(self):
            camera_move(self._camera_x, self._camera_y, self._camera_z, self._camera_rotate)
            renpy.restart_interaction()

        def layer_reset(self):
            global _3d_layers
            for layer in _3d_layers:
                layer_move(layer, self._3d_layers[layer])
            renpy.restart_interaction()

        def x_changed(self, v):
            camera_move(v - self.range_camera_pos, _camera_y, _camera_z, _camera_rotate)
            self.set_anchor_point()
            renpy.restart_interaction()

        def y_changed(self, v):
            camera_move(_camera_x, v - self.range_camera_pos, _camera_z, _camera_rotate)
            self.set_anchor_point()
            renpy.restart_interaction()

        def z_changed(self, v):
            camera_move(_camera_x, _camera_y, v - self.range_camera_pos, _camera_rotate)
            self.set_anchor_point()
            renpy.restart_interaction()

        def r_changed(self, v):
            camera_move(_camera_x, _camera_y, _camera_z, v - self.range_rotate)
            self.set_anchor_point()
            renpy.restart_interaction()

        def generate_layer_z_changed(self, l):
            def layer_z_changed(v):
                layer_move(l, v)
                renpy.restart_interaction()
            return layer_z_changed

        def put_clipboard(self, camera_tab, layer=""):
            string = '$camera_move(%s, %s, %s, %s, duration=0)' % (_camera_x, _camera_y, _camera_z, _camera_rotate)
            if not camera_tab:
                string = '$layer_move("%s", %s, duration=0)' % (layer, _3d_layers[layer])
            try:
                from pygame import scrap, locals
                scrap.put(locals.SCRAP_TEXT, string)
            except:
                renpy.notify(_("Can't open clipboard"))
            else:
                renpy.notify(__("Putted '%s' on clipboard") % string)

        def edit_value(self, function, range):
            v = renpy.invoke_in_new_context(renpy.call_screen, "_input_screen")
            if v:
                try:
                    function(renpy.python.py_eval(v) + range)
                except:
                    renpy.notify(_("Please type value"))

        def set_anchor_point(self):
            anchor_points = _viewers.all_anchor_points.get("camera", [])
            if anchor_points:
                for i, (t, c, w) in enumerate(anchor_points):
                    if _viewers.time < t:
                        anchor_points.insert(i, (_viewers.time, (_camera_x, _camera_y, _camera_z, _camera_rotate, _viewers.time, _viewers.warper), _viewers.warper))
                        break
                    elif _viewers.time == t:
                        anchor_points[i] = (_viewers.time, (_camera_x, _camera_y, _camera_z, _camera_rotate, _viewers.time, _viewers.warper), _viewers.warper)
                        break
                else:
                    anchor_points.append((_viewers.time, (_camera_x, _camera_y, _camera_z, _camera_rotate, _viewers.time, _viewers.warper), _viewers.warper))
            else:
                _viewers.all_anchor_points["camera"] = [(_viewers.time, (_camera_x, _camera_y, _camera_z, _camera_rotate, _viewers.time, _viewers.warper), _viewers.warper)]
            _viewers.sort_anchor_points()

        def play(self, play):
            check_points = []
            if "camera" in _viewers.all_anchor_points:
                for t, c, w in _viewers.all_anchor_points["camera"]:
                    check_points.append(c)
                camera_moves(check_points, play=play)

    _camera_viewer = CameraViewer()

    def _position_viewer():
        if not config.developer:
            return
        _viewers.clear_anchor_points()
        _viewers.time = 0
        _transform_viewer.init()
        _camera_viewer.init()
        renpy.invoke_in_new_context(renpy.call_screen, "_position_viewer")
        _camera_viewer.layer_reset()
        _camera_viewer.camera_reset()

screen _rot(): #show rule of thirds
    for i in range(1, 3):
        add Solid("#F00", xsize=config.screen_width, ysize=1, ypos=config.screen_height*i/3) 
        add Solid("#F00", xsize=1, ysize=config.screen_height, xpos=config.screen_width*i/3)

transform _delay_show(time):
    alpha 0
    pause time
    alpha 1

screen _position_viewer(tab="images", layer="master", tag="", time=0):
    if time:
        timer time+1 action Function(_viewers.change_time, _viewers.time)
    key "game_menu" action Return()
    zorder 10

    frame:
        if time:
            at _delay_show(time + 1)
        xfill True
        style_group "position_viewer"
        vbox:

            hbox:
                xfill False
                textbutton _("Images") action [SelectedIf(tab == "images"), Show("_position_viewer", tab="images")]
                textbutton _("3D Layers") action [SelectedIf(tab == "3D"), Show("_position_viewer", tab="3D")]
                textbutton _("Camera") action [SelectedIf(tab == "camera"), Show("_position_viewer", tab="camera")]
                textbutton _("Rule of thirds") action [SelectedIf(renpy.get_screen("_rot")), If(renpy.get_screen("_rot"), true=Hide("_rot"), false=Show("_rot"))]
            null height 10
            if tab == "images":
                hbox:
                    xfill False
                    label _("layers")
                    for l in config.layers:
                        if l not in ["screens", "transient", "overlay"]:
                            textbutton "[l]" action [SelectedIf(l == layer), Show("_position_viewer", tab=tab, layer=l)]
                hbox:
                    xfill False
                    label _("images")
                    for t in _transform_viewer.state_org[layer]:
                        textbutton "[t]" action [SelectedIf(t == tag), Show("_position_viewer", tab=tab, layer=layer, tag=t)]

                if tag:
                    textbutton _("clip board") action Function(_transform_viewer.put_show_clipboard, tag, layer)
                    for p in sorted(_transform_viewer.props_default.keys()):
                        $state = _transform_viewer.get_state(layer, tag, p)
                        if isinstance(state, int):
                            hbox:
                                $ f = _transform_viewer.generate_changed(layer, tag, p, True)
                                textbutton "[p]" action Function(_transform_viewer.put_prop_clipboard, p, state)
                                textbutton "[state]" action Function(_transform_viewer.edit_value, f, int)
                                bar adjustment ui.adjustment(range=_transform_viewer.int_range*2, value=state+_transform_viewer.int_range, page=1, changed=f) xalign 1.
                        elif isinstance(state, float):
                            hbox:
                                $ f = _transform_viewer.generate_changed(layer, tag, p)
                                textbutton "[p]" action Function(_transform_viewer.put_prop_clipboard, p, state)
                                textbutton "[state:>.4]" action Function(_transform_viewer.edit_value, f)
                                bar adjustment ui.adjustment(range=_transform_viewer.float_range*2, value=state+_transform_viewer.float_range, page=.05, changed=f) xalign 1.

            elif tab == "camera":
                textbutton _("clip board") action Function(_camera_viewer.put_clipboard, True)
                hbox:
                    label "x"
                    textbutton "[_camera_x]" action Function(_camera_viewer.edit_value, _camera_viewer.x_changed, _camera_viewer.range_camera_pos)
                    bar adjustment ui.adjustment(range=_camera_viewer.range_camera_pos*2, value=_camera_x+_camera_viewer.range_camera_pos, page=1, changed=_camera_viewer.x_changed) xalign 1.
                hbox:
                    label "y"
                    textbutton "[_camera_y]" action Function(_camera_viewer.edit_value, _camera_viewer.y_changed, _camera_viewer.range_camera_pos)
                    bar adjustment ui.adjustment(range=_camera_viewer.range_camera_pos*2, value=_camera_y+_camera_viewer.range_camera_pos, page=1, changed=_camera_viewer.y_changed) xalign 1.
                hbox:
                    label "z"
                    textbutton "[_camera_z]" action Function(_camera_viewer.edit_value, _camera_viewer.z_changed, _camera_viewer.range_camera_pos)
                    bar adjustment ui.adjustment(range=_camera_viewer.range_camera_pos*2, value=_camera_z+_camera_viewer.range_camera_pos, page=1, changed=_camera_viewer.z_changed) xalign 1.
                hbox:
                    label "rotate"
                    textbutton "[_camera_rotate]" action Function(_camera_viewer.edit_value, _camera_viewer.r_changed, _camera_viewer.range_rotate)
                    bar adjustment ui.adjustment(range=_camera_viewer.range_rotate*2, value=_camera_rotate+_camera_viewer.range_rotate, page=1, changed=_camera_viewer.r_changed) xalign 1.
            elif tab == "3D":
                for layer in sorted(_3d_layers.keys()):
                    hbox:
                        textbutton "[layer]" action Function(_camera_viewer.put_clipboard, False, layer)
                        textbutton "{}".format(int(_3d_layers[layer])) action Function(_camera_viewer.edit_value, _camera_viewer.generate_layer_z_changed(layer), 0)
                        bar adjustment ui.adjustment(range=_camera_viewer.range_layer_z, value=_3d_layers[layer], page=1, changed=_camera_viewer.generate_layer_z_changed(layer)) xalign 1.
            hbox:
                xfill False
                xalign 1.
                if tab == "images":
                    textbutton _("reset") action [_transform_viewer.reset, renpy.restart_interaction]
                elif tab == "camera":
                    textbutton _("reset") action [_camera_viewer.camera_reset, renpy.restart_interaction]
                elif tab == "3D":
                    textbutton _("reset") action [_camera_viewer.layer_reset, renpy.restart_interaction]
                textbutton _("close") action Return()

    frame:
        if time:
            at _delay_show(time + 1)
        xfill True
        yalign 1.0
        vbox:

            hbox:
                textbutton _("time [_viewers.time:>.2f]s") action Function(_viewers.edit_time)
                textbutton _("warper") action _viewers.select_time_warper
                textbutton _("<") action Function(_viewers.prev_time)
                textbutton _(">") action Function(_viewers.next_time)
                bar adjustment ui.adjustment(range=7.0, value=_viewers.time, changed=_viewers.change_time) xalign 1.
            hbox:
                style_group "position_viewer"
                textbutton _("clear anchor points") action [Function(_viewers.clear_anchor_points), Function(_camera_viewer.set_anchor_point)]
                textbutton _("remove anchor point") action [Function(_viewers.remove_anchor_point, _viewers.time)]
                textbutton _("play") action [SensitiveIf(_viewers.sorted_anchor_points[-1] > 0), Function(_camera_viewer.play, play=True), Function(_transform_viewer.play, play=True), Hide("_position_viewer"), Show("_position_viewer", tab=tab, layer=layer, tag=tag, time=_viewers.sorted_anchor_points[-1]), renpy.restart_interaction]
                textbutton _("clipboard") action Function(_viewers.put_clipboard)

init -1600:
    style position_viewer_frame background "#0006"
    style position_viewer_button size_group "position_viewer"
    style position_viewer_button_text xalign .5
    style position_viewer_label xminimum 100
    style position_viewer_vbox xfill True

screen _input_screen(message="type value", default=""):
    modal True
    zorder 100
    key "game_menu" action Return("")

    frame:
        style_group "input_screen"

        has vbox

        label message

        hbox:
            input default default

init -1600:
    style input_screen_frame xfill True ypos .1 xmargin .05 ymargin .05
    style input_screen_vbox xfill True spacing 30
    style input_screen_label xalign .5
    style input_screen_hbox  xalign .5

screen _warper_selecter():
    modal True
    zorder 100
    key "game_menu" action Return("")

    frame:
        yalign .98
        style_group "warper_selecter"

        has vbox

        label _("Select a warper function")

        hbox:
            for warper in renpy.atl.warpers:
                textbutton warper action [SetField(_viewers, "warper", warper), Return()]


init -1098 python:
    # overwrite keymap
    km = renpy.Keymap(
        rollback = renpy.rollback,
        screenshot = _screenshot,
        toggle_fullscreen = renpy.toggle_fullscreen,
        toggle_skip = _keymap_toggle_skipping,
        fast_skip = _fast_skip,
        game_menu = _invoke_game_menu,
        hide_windows = renpy.curried_call_in_new_context("_hide_windows"),
        launch_editor = _launch_editor,
        reload_game = _reload_game,
        developer = _developer,
        quit = renpy.quit_event,
        iconify = renpy.iconify,
        help = _help,
        choose_renderer = renpy.curried_call_in_new_context("_choose_renderer"),
        console = _console.enter,
        profile_once = _profile_once,
        self_voicing = Preference("self voicing", "toggle"),
        position_viewer = _position_viewer,
        )

    config.underlay = [ km ]

    del km


init 1100 python:
    config.locked = False
    config.keymap["position_viewer"] = ['P']
    config.locked = True
